from lxml import etree

from xml.sax.saxutils import escape

xml_encoding = '<?xml version="1.0" encoding="UTF-8"?>\n'
METADATA_NAMESPACE = "http://soap.sforce.com/2006/04/metadata"


def serialize_xml_for_salesforce(xdoc, xml_declaration=True):
    r = xml_encoding if xml_declaration else ""
    if METADATA_NAMESPACE in xdoc.getroot().tag:
        xdoc.getroot().attrib["xmlns"] = METADATA_NAMESPACE
    for action, elem in etree.iterwalk(
        xdoc, events=("start", "end", "start-ns", "end-ns", "comment")
    ):
        if action == "start-ns":
            prefix, ns = elem
            if ns != METADATA_NAMESPACE:
                raise AssertionError(
                    f"The only namespace that is supported is {METADATA_NAMESPACE}, not {ns}"
                )
        elif action == "start":
            tag = elem.tag
            if "}" in tag:
                tag = tag.split("}")[1]
            text = (
                escape(elem.text, {"'": "&apos;", '"': "&quot;"})
                if elem.text is not None
                else ""
            )

            attrs = "".join([f' {k}="{v}"' for k, v in elem.attrib.items()])
            if not has_content(elem):
                r += f"<{tag}{attrs}/>"
            else:
                r += f"<{tag}{attrs}>{text}"
        elif action == "end" and has_content(elem):
            tag = elem.tag
            if "}" in tag:
                tag = tag.split("}")[1]
            tail = elem.tail if elem.tail else "\n"
            r += f"</{tag}>{tail}"
        elif action == "comment":
            r += str(elem) + (elem.tail if elem.tail else "")
    return r


def has_content(element):
    return element.text or list(element)



from typing import Dict, List, Union, IO
from logging import getLogger
from pathlib import Path

from pydantic import Field, validator, ValidationError

from cumulusci.utils.yaml.model_parser import CCIDictModel

LOGGER_NAME = "MAPPING_LOADER"
logger = getLogger(LOGGER_NAME)


class MappingLookup(CCIDictModel):
    "Lookup relationship between two tables."
    table: str
    key_field: str = None
    value_field: str = None
    join_field: str = None
    after: str = None
    aliased_table: str = None


class MappingStep(CCIDictModel):
    "Step in a load or extract process"
    sf_object: str
    table: str = None
    fields_: Dict[str, str] = Field(..., alias="fields")
    lookups: Dict[str, MappingLookup] = {}
    static: Dict[str, str] = {}
    filters: List[str] = []
    action: str = "insert"
    oid_as_pk: bool = False  # this one should be discussed and probably deprecated
    record_type: str = None  # should be discussed and probably deprecated

    @validator("record_type")
    def record_type_is_deprecated(cls, v):
        logger.warning(
            "record_type is deprecated. Just supply a RecordTypeId column declaration and it will be inferred"
        )
        return v

    @validator("oid_as_pk")
    def oid_as_pk_is_deprecated(cls, v):
        logger.warning(
            "oid_as_pk is deprecated. Just supply an Id column declaration and it will be inferred."
        )
        return v


class MappingSteps(CCIDictModel):
    "Mapping of named steps"
    __root__: Dict[str, MappingStep]


ValidationError = ValidationError  # export Pydantic's Validation Error under an alias


def parse_from_yaml(source: Union[str, Path, IO]) -> Dict:
    "Parse from a path, url, path-like or file-like"
    return MappingSteps.parse_from_yaml(source)



from glob import glob
from pathlib import Path
from tempfile import TemporaryDirectory
from io import StringIO

from lxml import etree
import pytest

from cumulusci.utils.xml.salesforce_encoding import serialize_xml_for_salesforce


class TestSalesforceEncoding:
    def test_xml_declaration(self):
        xml = etree.parse(StringIO("<foo/>"))
        print("xml", xml)
        out = serialize_xml_for_salesforce(xml, xml_declaration=True)
        assert out.startswith("<?xml")

        out = serialize_xml_for_salesforce(xml, xml_declaration=False)
        assert not out.startswith("<?xml")

        out = serialize_xml_for_salesforce(xml)
        assert out.startswith("<?xml")

    def test_roundtripping(self):
        from cumulusci.tasks.metadata import tests

        print("Q", Path(tests.__file__).parent / "sample_package.xml")

        print("r", Path(tests.__file__).parent / "sample_package.xml")

        files = glob(str(Path(tests.__file__).parent / "sample_package.xml"))

        #   If you fiddle with the salesforce encoder, the code below may be
        #   useful to ensure that it faithfully round-trips, but it only works
        #   if run in a directory with a parent-directory which contains a few
        #   CCI projects which contain translations in them. At the time of
        #   writing it roundtripped 421 files without a byte of difference.
        #
        # files = glob("../*/*/*ranslations/**ranslation", recursive=True)

        assert files

        for file in files:
            orig = open(file).read()
            tree = etree.parse(file)
            out = serialize_xml_for_salesforce(tree)
            print(f"[{orig}]")
            print(f"[{out}]")
            assert orig == out, f"Filename did not roundtrip cleanly {file}"
            print("PASSED", file)

    def tests_filename_bad_xml(self):
        with TemporaryDirectory() as d:
            filename = d + "/foobar.notxml"
            with open(filename, "w") as f:
                f.write("<<<<")
            with pytest.raises(SyntaxError) as e:
                etree.parse(filename)
            assert "foobar.notxml" in str(e.value)

    def test_comment(self):
        tree = etree.parse(
            StringIO(
                "<Foo><!-- Salesforce files should not have comments, but just in case! --> </Foo>"
            )
        )
        assert "just in case! -->" in serialize_xml_for_salesforce(tree)

    def test_namespaces(self):
        tree = etree.parse(StringIO("<Foo xmlns:foo='https://html5zombo.com/'></Foo>"))
        with pytest.raises(AssertionError):
            serialize_xml_for_salesforce(tree)
